import re
import html
import requests
import unicodedata
import pandas as pd
import xml.etree.ElementTree as ET

url = "https://sdmx.oecd.org/public/rest/dataflow/all"

NS = {
    "message": "http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message",
    "structure": "http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure",
    "common": "http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common",
    "xml": "http://www.w3.org/XML/1998/namespace",
}

_ws_re = re.compile(r"\s+")
_tag_re = re.compile(r"<[^>]+>")

def _clean(s: str | None) -> str | None:
    if s is None: return None
    s = html.unescape(s)
    s = _tag_re.sub("", s)
    s = _ws_re.sub(" ", s).strip()
    return s or None

# Cache
_datasets: pd.DataFrame | None = None

def _load_datasets() -> pd.DataFrame:
    """
    Loads OECD datasets and keeps them in memory.
    """
    global _datasets
    if _datasets is not None: return _datasets

    headers = {"Accept": "application/vnd.sdmx.structure+xml;version=2.1"}
    r = requests.get(url, headers=headers, timeout=30)
    r.raise_for_status()
    root = ET.fromstring(r.content)

    rows = []
    for df in root.findall(".//structure:Dataflow", NS):
        dataflow_id = df.attrib.get("id")
        agency_id = df.attrib.get("agencyID")

        name_elem = df.find("common:Name[@xml:lang='en']", NS)
        desc_elem = df.find("common:Description[@xml:lang='en']", NS)

        name = _clean("".join(name_elem.itertext())) if name_elem is not None else None
        desc_raw = "".join(desc_elem.itertext()) if desc_elem is not None else None
        desc = _clean(desc_raw)

        rows.append(
            {
                "dataflowID": dataflow_id,
                "agencyID": agency_id,
                "name": name,
                "description": desc,
            }
        )

    _datasets = pd.DataFrame(rows)
    return _datasets

def search_keywords(*keywords: str) -> pd.DataFrame:
    """
    Searches OECD datasets for a set of keywords.

    Args:
        keywords (str | list[str]): Single keyword or list of keywords. Acts as OR.

    Returns:
        pd.DataFrame: Matching rows.
    """
    datasets = _load_datasets()

    # Clean and validate keywords
    keywords = [k for k in keywords if isinstance(k, str) and k.strip()]
    if not keywords:
        raise ValueError("No valid keywords provided.")

    def _normalize_series(s: pd.Series) -> pd.Series:
        s = s.fillna("").astype(str).str.lower()
        return s.map(
            lambda x: "".join(
                ch for ch in unicodedata.normalize("NFKD", x)
                if not unicodedata.combining(ch)
            )
        )

    # Combined normalized text for each row
    text = (
        datasets["name"].fillna("").astype(str)
        + " "
        + datasets["description"].fillna("").astype(str)
    )
    text_norm = _normalize_series(text)

    def _normalize_kw(kw: str) -> str:
        kw = unicodedata.normalize("NFKD", kw.lower())
        return "".join(ch for ch in kw if not unicodedata.combining(ch))

    norm_keywords = [_normalize_kw(k) for k in keywords]

    overall_mask = pd.Series(False, index=datasets.index)
    score = pd.Series(0, index=datasets.index, dtype="int64")

    for kw in norm_keywords:
        m = text_norm.str.contains(kw, na=False, regex=False)
        overall_mask |= m
        score = score.add(m.astype("int8"), fill_value=0)

    result = datasets.loc[overall_mask].copy()
    result["_match_score"] = score.loc[overall_mask]
    result = result.sort_values("_match_score", ascending=False)

    return result[["agencyID", "dataflowID", "name", "description"]]


import importlib
import requests
import pandas as pd


def _fake_structure_xml() -> bytes:
    return  b"""<?xml version="1.0" encoding="UTF-8"?>
                <message:Structure
                xmlns:message="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message"
                xmlns:structure="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure"
                xmlns:common="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common"
                xmlns:xml="http://www.w3.org/XML/1998/namespace"
                >
                <message:Structures>

                    <structure:Concepts>
                    <structure:ConceptScheme>
                        <structure:Concept id="PRICES">
                        <common:Name xml:lang="en">Prices</common:Name>
                        <structure:CoreRepresentation>
                            <structure:Enumeration>
                            <Ref id="CL_PRICES"/>
                            </structure:Enumeration>
                        </structure:CoreRepresentation>
                        </structure:Concept>
                    </structure:ConceptScheme>
                    </structure:Concepts>

                    <structure:Codelists>
                    <structure:Codelist id="CL_PRICES">
                        <structure:Code id="Q">
                        <common:Name xml:lang="en">Quarterly</common:Name>
                        </structure:Code>
                        <structure:Code id="V">
                        <common:Name xml:lang="en">Volume</common:Name>
                        </structure:Code>
                    </structure:Codelist>
                    </structure:Codelists>

                    <structure:Constraints>
                    <structure:ContentConstraint>
                        <structure:CubeRegion>
                        <common:KeyValue id="PRICES">
                            <common:Value>Q</common:Value>
                            <common:Value>V</common:Value>
                        </common:KeyValue>
                        </structure:CubeRegion>
                    </structure:ContentConstraint>
                    </structure:Constraints>

                    <structure:DataStructures>
                    <structure:DataStructure>
                        <structure:DataStructureComponents>
                        <structure:DimensionList>
                            <structure:Dimension id="PRICES" position="1"/>
                        </structure:DimensionList>
                        </structure:DataStructureComponents>
                    </structure:DataStructure>
                    </structure:DataStructures>

                </message:Structures>
                </message:Structure>
                """


class _Resp:
    def __init__(self, content: bytes, status_code: int = 200):
        self.content = content
        self.status_code = status_code


def test_get_structure_builds_toc_values_and_explain(monkeypatch):
    def fake_get(url, *args, **kwargs):
        if "/public/rest/dataflow/" in url and "?references=all" in url:
            return _Resp(_fake_structure_xml())
        raise AssertionError(f"Unexpected URL in test_structure: {url}")

    # Patch the requests used by notoecd.structure
    monkeypatch.setattr(requests, "get", fake_get)

    structure_mod = importlib.import_module("notoecd.structure")
    importlib.reload(structure_mod)

    # Clear cache so test is isolated
    structure_mod.get_structure.cache_clear()

    s = structure_mod.get_structure("OECD.CFE.EDS", "DSD_REG_ECO@DF_GDP")

    assert isinstance(s.toc, pd.DataFrame)
    assert list(s.toc["title"]) == ["PRICES"]
    assert s.toc.loc[0, "values"] == ["Q", "V"]

    assert isinstance(s.concepts, dict)
    assert "CODELISTS" in s.concepts
    assert "PRICES" in s.concepts

    d = s.explain_vals("PRICES")
    assert d == {"Q": "Quarterly", "V": "Volume"}



import requests
import pandas as pd
from functools import lru_cache
import xml.etree.ElementTree as ET

class Structure():
    def __init__(self, agencyID:str, dataflowID:str):
        self.ns = {
            'message': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message',
            'common': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common',
            'structure': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure',
            'xml': 'http://www.w3.org/XML/1998/namespace'
        }

        url = f"https://sdmx.oecd.org/public/rest/dataflow/{agencyID}/{dataflowID}/?references=all"
        response = requests.get(url)
        tree = ET.ElementTree(ET.fromstring(response.content))
        self.root = tree.getroot()

        self.concepts = self.build_concepts_dict()
        self.values = self.build_values_dict()
        self.toc = self.build_toc()
        self.params = list(self.values.keys())

    def build_concepts_dict(self) -> dict:
        # Extract dimensions from the Concepts section
        concepts = {'DIMENSIONS':{}, 'CODELISTS':{}}
        conceptScheme_section = self.root.findall('.//message:Structures/structure:Concepts/structure:ConceptScheme', self.ns)
        if conceptScheme_section:
            for concept_scheme in conceptScheme_section:
                for concept in concept_scheme.findall('structure:Concept', self.ns):
                    concept_id = concept.get('id')
                    name_elem = concept.find('common:Name[@{http://www.w3.org/XML/1998/namespace}lang="en"]', self.ns)
                    if concept_id and name_elem is not None:
                        concepts['DIMENSIONS'][concept_id] = name_elem.text
                    for core_representation in concept.findall('structure:CoreRepresentation', self.ns):
                        for enumeration in core_representation.findall('structure:Enumeration', self.ns):
                            ref_elem = enumeration.find('Ref')
                            if ref_elem is not None:
                                codelist_id = ref_elem.get('id')
                                if concept_id and codelist_id:
                                    concepts['CODELISTS'][concept_id] = (
                                        codelist_id.split("CL_")[1] if codelist_id.startswith("CL_") 
                                        else codelist_id
                                    )
        
        # Extract codes from the Codelists section
        codes_section = self.root.findall('.//message:Structures/structure:Codelists/structure:Codelist', self.ns)
        if codes_section:
            for code_scheme in codes_section:
                dimension = code_scheme.get('id')[3:]
                concepts[dimension] = {}
                for code in code_scheme.findall('structure:Code', self.ns):
                    code_id = code.get('id')
                    name_elem = code.find('common:Name[@{http://www.w3.org/XML/1998/namespace}lang="en"]', self.ns)
                    if code_id and name_elem is not None:
                        concepts[dimension][code_id] = name_elem.text
        return concepts

    def build_values_dict(self) -> dict:
        # Extract values from Constraints section
        values = {}
        constraints_section = self.root.findall('.//message:Structures/structure:Constraints/structure:ContentConstraint', self.ns)
        if constraints_section:
            constraint = constraints_section[0]
            cube_region = constraint.find('structure:CubeRegion', self.ns)
            if cube_region is not None:
                for key_value in cube_region.findall('common:KeyValue', self.ns):
                    key = key_value.get('id')
                    value_elems = key_value.findall('common:Value', self.ns)
                    if key is not None and value_elems:
                        values[key] = [value_elem.text for value_elem in value_elems if value_elem.text]
        return values

    def build_toc(self) -> pd.DataFrame:
        # Builds a DataFrame containing the call dimensions and options (table of contents)
        data_structure = self.root.find('.//message:Structures/structure:DataStructures/structure:DataStructure', self.ns)
        dimension_list = data_structure.find('.//structure:DataStructureComponents/structure:DimensionList', self.ns)
        rows = []
        
        if dimension_list is not None:
            dimensions = dimension_list.findall('.//structure:Dimension', self.ns)
            
            # Sort dimensions by their 'position' to ensure correct order
            sorted_dimensions = sorted(dimensions, key=lambda dim: int(dim.get('position')))
            
            for dimension in sorted_dimensions:
                dimension_id = dimension.get('id')
                position = dimension.get('position')

                # Get the corresponding values for the dimension (if present in the 'values' dictionary)
                dimension_values = self.values.get(dimension_id, [])
                
                rows.append({
                    'id': int(position),
                    'title': dimension_id,
                    'values': dimension_values
                })

        df = pd.DataFrame(rows)
        df = df.sort_values(by='id').reset_index(drop=True)
        return df

    def explain_vals(self, dimension:str) -> dict:
        # Explains the parameters of a given dimension
        dimension = dimension.strip().upper()
        concepts = self.concepts
        values = self.values[dimension]
        codelists = concepts['CODELISTS']

        def fallback(dimension, concepts):
            # Look for subset of keys
            toc = self.toc
            concept_mapping = {}
            for title in toc.title:
                vals = set(toc[toc['title'] == title].values[0][2])
                concept_mapping[title] = [i for i in concepts.keys() if vals.issubset(set(concepts[i].keys()))][0]
            concept_mapping['FREQ'] = 'FREQ'
            if (dimension not in concepts) and (dimension in concept_mapping): 
                    dimension = concept_mapping[dimension]
            concepts = concepts[dimension]
            return {i: concepts[i] for i in values}

        if dimension not in codelists:
            if dimension in concepts: 
                return concepts[dimension]
            return fallback(dimension, concepts)
        
        dim_codelist = codelists[dimension]
        explanation = concepts[dim_codelist]
        clean_values = [i for i in values if i in explanation]

        if len(clean_values) == 0:
            return fallback(dimension, concepts)

        unclean_values = [i for i in values if i not in explanation]
        if len(unclean_values) > 0:
            print(f"Could not find explanation for {unclean_values}")

        return {i:explanation[i] for i in clean_values}

@lru_cache(maxsize=128)
def get_structure(agencyID: str, dataflowID: str) -> Structure:
    return Structure(agencyID, dataflowID)



import notoecd

def test_public_api_exports():
    assert callable(notoecd.get_df)
    assert callable(notoecd.get_structure)
    assert callable(notoecd.search_keywords)

def test_import_package():
    import importlib
    m = importlib.import_module("notoecd")
    assert m is not None



import importlib
import requests


def _fake_dataflow_all_xml() -> bytes:
    xml = """<?xml version="1.0" encoding="UTF-8"?>
            <message:Structure
            xmlns:message="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message"
            xmlns:structure="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure"
            xmlns:common="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common"
            xmlns:xml="http://www.w3.org/XML/1998/namespace"
            >
            <message:Structures>
                <structure:Dataflows>
                <structure:Dataflow id="DSD_REG_ECO@DF_GDP" agencyID="OECD.CFE.EDS">
                    <common:Name xml:lang="en">Gross domestic product - Regions</common:Name>
                    <common:Description xml:lang="en">GDP by region</common:Description>
                </structure:Dataflow>

                <structure:Dataflow id="DF_CAFE" agencyID="OECD.TEST">
                    <common:Name xml:lang="en">Café prices</common:Name>
                    <common:Description xml:lang="en">Prices in cafes</common:Description>
                </structure:Dataflow>

                <structure:Dataflow id="DF_OTHER" agencyID="OECD">
                    <common:Name xml:lang="en">Other dataset</common:Name>
                    <common:Description xml:lang="en">Other description</common:Description>
                </structure:Dataflow>
                </structure:Dataflows>
            </message:Structures>
            </message:Structure>
            """
    return xml.encode("utf-8")


class _Resp:
    def __init__(self, content: bytes, status_code: int = 200):
        self.content = content
        self.status_code = status_code

    def raise_for_status(self):
        if self.status_code >= 400:
            raise requests.HTTPError(f"HTTP {self.status_code}")


def test_datasets_lazy_loaded_and_cached_in_memory(monkeypatch):
    calls = {"n": 0}

    def fake_get(url, *args, **kwargs):
        if url.endswith("/public/rest/dataflow/all"):
            calls["n"] += 1
            return _Resp(_fake_dataflow_all_xml())
        raise AssertionError(f"Unexpected URL in test_datasets: {url}")

    monkeypatch.setattr(requests, "get", fake_get)

    datasets_mod = importlib.import_module("notoecd.datasets")
    importlib.reload(datasets_mod)

    assert calls["n"] == 0

    hits = datasets_mod.search_keywords("gdp")
    assert calls["n"] == 1
    assert len(hits) == 1

    hits2 = datasets_mod.search_keywords("cafe")
    assert calls["n"] == 1
    assert len(hits2) == 1


def test_search_keywords_or_and_normalization(monkeypatch):
    def fake_get(url, *args, **kwargs):
        if url.endswith("/public/rest/dataflow/all"):
            return _Resp(_fake_dataflow_all_xml())
        raise AssertionError(f"Unexpected URL in test_datasets: {url}")

    monkeypatch.setattr(requests, "get", fake_get)

    datasets_mod = importlib.import_module("notoecd.datasets")
    importlib.reload(datasets_mod)

    hits = datasets_mod.search_keywords("gross domestic product", "cafe")

    assert len(hits) == 2
    assert any(hits["dataflowID"] == "DSD_REG_ECO@DF_GDP")
    assert any(hits["dataflowID"] == "DF_CAFE")

    names = " ".join(hits["name"].fillna("").tolist()).lower()
    assert ("gross domestic product" in names) or ("café" in names) or ("cafe" in names)


def test_search_keywords_rejects_empty(monkeypatch):
    def fake_get(url, *args, **kwargs):
        if url.endswith("/public/rest/dataflow/all"):
            return _Resp(_fake_dataflow_all_xml())
        raise AssertionError(f"Unexpected URL in test_datasets: {url}")

    monkeypatch.setattr(requests, "get", fake_get)

    datasets_mod = importlib.import_module("notoecd.datasets")
    importlib.reload(datasets_mod)

    try:
        datasets_mod.search_keywords("   ", "")
        raise AssertionError("Expected ValueError for empty keywords")
    except ValueError:
        pass


import pandas as pd
from types import SimpleNamespace
from unittest.mock import patch
import notoecd.calls as calls


def _fake_structure_with_toc_titles(titles):
    toc = pd.DataFrame({"title": titles})
    return SimpleNamespace(toc=toc)


def test_build_filter_expression_orders_by_toc_and_uppercases():
    fake_s = _fake_structure_with_toc_titles(["PRICES", "UNIT_MEASURE", "MEASURE"])
    filters = {"prices": "q", "unit_measure": ["USD_PPP_PS"], "measure": "gdp"}

    with patch("notoecd.calls.get_structure", return_value=fake_s):
        expr = calls._build_filter_expression("A", "B", filters)

    assert expr == "Q.USD_PPP_PS.GDP"


def test_build_filter_expression_missing_dims_are_empty_parts():
    fake_s = _fake_structure_with_toc_titles(["A", "B", "C"])

    with patch("notoecd.calls.get_structure", return_value=fake_s):
        expr = calls._build_filter_expression("A", "B", {"b": "x"})

    assert expr == ".X."


def test_build_filter_expression_multi_value_joins_plus():
    fake_s = _fake_structure_with_toc_titles(["territorial_level"])

    with patch("notoecd.calls.get_structure", return_value=fake_s):
        expr = calls._build_filter_expression("A", "B", {"territorial_level": ["tl2", "tl3"]})

    assert expr == "TL2+TL3"


def test_get_df_builds_url_and_returns_copy():
    calls._fetch_df.cache_clear()

    fake_s = _fake_structure_with_toc_titles(["PRICES"])
    fake_df = pd.DataFrame({"x": [1, 2]})

    with patch("notoecd.calls.get_structure", return_value=fake_s), \
         patch("notoecd.calls.pd.read_csv", return_value=fake_df) as mock_read_csv:
        out = calls.get_df("OECD.CFE.EDS", "DSD_REG_ECO@DF_GDP", {"prices": "q"})

    assert out.equals(fake_df)
    assert out is not fake_df  # must be a copy()

    (url,), kwargs = mock_read_csv.call_args
    assert url.startswith("https://sdmx.oecd.org/public/rest/data/")
    assert "OECD.CFE.EDS,DSD_REG_ECO@DF_GDP," in url
    assert "/Q" in url
    assert "dimensionAtObservation=AllDimensions" in url
    assert "format=csvfile" in url
    assert kwargs["storage_options"]["User-Agent"]


def test_get_df_accepts_string_filter_expression_and_uppercases():
    calls._fetch_df.cache_clear()

    fake_df = pd.DataFrame({"x": [1]})
    with patch("notoecd.calls.pd.read_csv", return_value=fake_df) as mock_read_csv:
        _ = calls.get_df("A", "B", " tl2+tl3..gdp ")

    (url,), _ = mock_read_csv.call_args
    assert "/TL2+TL3..GDP" in url



import pandas as pd
from typing import Union
from functools import lru_cache
from .structure import get_structure

@lru_cache(maxsize=64)
def _fetch_df(url: str) -> pd.DataFrame:
    return pd.read_csv(url, storage_options={"User-Agent": "Mozilla/5.0"})


def _clean(s: str) -> str: 
    return str(s).strip().lower()


def _build_filter_expression(
    agencyID: str,
    dataflowID: str,
    filters: dict,
) -> str:
    
    s = get_structure(agencyID, dataflowID)
    filters = {_clean(k): v for k, v in filters.items()}

    parts = []
    for dim in s.toc.title:
        dim_key = _clean(dim)
        if dim_key in filters:
            val = filters[dim_key]
            if isinstance(val, str):
                val = [val]
            parts.append("+".join(_clean(v) for v in val))
        else:
            parts.append("")

    return ".".join(parts).upper()


def get_df(
    agencyID: str,
    dataflowID: str,
    filters: Union[str, dict] = "",
    version: str = "",
) -> pd.DataFrame:
    
    if isinstance(filters, dict):
        filter_expression = _build_filter_expression(agencyID, dataflowID, filters)
    else:
        filter_expression = _clean(filters).upper()

    url = (
        f"https://sdmx.oecd.org/public/rest/data/"
        f"{agencyID},{dataflowID},{version}/{filter_expression}"
        "?dimensionAtObservation=AllDimensions&format=csvfile"
    )

    base_df = _fetch_df(url)
    return base_df.copy()
